/*
 * acompare.cpp:
 *
 * Compare the contents of an ISO file to an AFF file.
 * Optionally, if they are equal, delete the ISO file
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */


#include "afflib.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <zlib.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <assert.h>

char *progname = "acompare";

int opt_remove = 0;
int opt_quiet  = 0;
int opt_batch  = 0;
char *batch_ext = "";

void usage()
{
    printf("usage: %s [-rqb] file(s)\n",progname);
    printf("where file(s):  AFF-file ISO-file --- compare the two files and say if equal.\n");
    printf("       -b img file.aff               --- compare file.aff and file.img\n");
    printf("       -b img file1.aff file2.aff... --- compare file1.aff, file1.img, etc.\n");
    printf("       -r Remove the ISO file if the two are equal.\n");
    printf("       -q Quiet; no output except for errors.\n");
    exit(0);
}



void compare_files(const char *aff_file,const char *iso_file)
{
     /* make sure the extension is right */
     char *cc = strrchr(aff_file,'.');
     if(!cc || strcmp(cc,".aff")!=0){
	 fprintf(stderr,"%s is not a valid AFF file.\n",aff_file);
	 exit(1);
     }

     /* See if they are the same file. That would be senseless */
     struct stat sa,si;

     if(stat(aff_file,&sa)) err(1,aff_file);
     if(stat(iso_file,&si)) err(1,iso_file);

     if(sa.st_dev==si.st_dev &&
	sa.st_ino==si.st_ino){
	 fprintf(stderr,"%s and %s are the same file!\n",aff_file,iso_file);
	 exit(1);
     }


     /* Kind of a hack. We use the af routines to open both files, since
      * they automatically fall-back to ISO files for non-AFF files.
      * In the process, we get to double-check the AFF routines as well...
      * (Don't worry; we'll also check the MD5.)
      */
     AFFILE *file1 = af_open(aff_file,O_RDONLY,0);
     if(!file1) err(1,"afopen %s",aff_file);
     
     AFFILE *file2 = af_open(iso_file,O_RDONLY,0);
     if(!file2) err(1,"afopen %s",iso_file);


     if(!opt_quiet){
	 printf("comparing %s and %s...\n",aff_file,iso_file);
     }

     /* Compare 1MB at a time, to minimize on seeks... */

     long long total_bytes = 0;
     struct timeval start;
     gettimeofday(&start,0);

     while(!af_eof(file1)){
	 char buf1[65536 * 16];
	 char buf2[65536 * 16];
	 size_t bytes1,bytes2;

	 if(!opt_quiet){
	     /* Calculate the elapsed time and print status information */
	     struct timeval now;
	    
	     gettimeofday(&now,0);
	     float percent_done = total_bytes*100.0/si.st_size;

	     printf("\r%qd out of %qd bytes compared  (%5.2f%%) ",
		    total_bytes,(long long)si.st_size,percent_done);

	     /* If we have non-zero time and non-zero bytes, report throughput */
	     long total_msec = (now.tv_sec-start.tv_sec)*1000 + (now.tv_usec-start.tv_usec)/1000;
	     long total_sec  = total_msec / 1000;

	     if(total_msec>0 && total_bytes>0){
		 printf("%6.2f MBytes/sec ",total_bytes/(total_msec * 1000.0));
	     }

	     /* Figure out how much time is left... */
	     if(percent_done>0){
		 long est_total_time = (long)(total_sec / (percent_done/100.0));
		 long est_time_remaining = est_total_time - total_sec;
		 printf("%2ld:%02ld ",est_time_remaining / 60, est_time_remaining % 60);
	     }
	     fflush(stdout);
	 }

	 bytes1 = af_read(file1,(unsigned char *)buf1,sizeof(buf1));
	 bytes2 = af_read(file2,(unsigned char *)buf2,sizeof(buf2));

	 if(bytes1 != bytes2){
	     fprintf(stderr,"\n");
	     fprintf(stderr,"Error at offset %qd. Attempted to read %lu bytes.\n",
		     total_bytes,(long unsigned int)sizeof(buf1));
	     fprintf(stderr,"afread(%s) = %lu bytes\n",aff_file,(long unsigned int)bytes1);
	     fprintf(stderr,"afread(%s) = %lu bytes\n",iso_file,(long unsigned int)bytes2);
	     exit(2);
	 }

	 if(memcmp(buf1,buf2,bytes1)!=0){
	     fprintf(stderr,"\n");
	     fprintf(stderr,"Error at offset %qd. Read %lu bytes.\n",
		     total_bytes,(long unsigned int)sizeof(buf1));
	     fprintf(stderr,"Files do not match.\n");
	     exit(3);
	 }
	 total_bytes += bytes1;
     }
     af_close(file1);
     af_close(file2);

     if(!opt_quiet){
	 printf("%60s\n","");
	 printf("Read %qd bytes. Files match!\n",total_bytes);
     }
}

int main(int argc,char **argv)
{
    int bflag, ch;

     progname = argv[0];
     bflag = 0;
     while ((ch = getopt(argc, argv, "s:h?rqb:")) != -1) {
             switch (ch) {
	     case 'r':
		 opt_remove++;
		 break;
	     case 'q':
		 opt_quiet++;
		 break;
	     case 'b':
		 opt_batch++;
		 batch_ext=optarg;
		 break;
	     case 'h':
             case '?':
             default:
		 usage();
             }
     }
     argc -= optind;
     argv += optind;

     if(opt_batch) while(argc){
	 char *aff_file = *argv++;
	 argc--;
	 char *iso_file = (char *)malloc(strlen(aff_file)+16);
	 char *cc;

	 strcpy(iso_file,aff_file);
	 cc = rindex(iso_file,'.');
	 if(cc){
	     strcpy(cc+1,batch_ext);
	 }
	 else {
	     strcat(iso_file,".");
	     strcat(iso_file,batch_ext);
	 }
	 compare_files(aff_file,iso_file);
	 free(iso_file);
	 exit(0);
     }

     if(argc==2){			// if just 2, compare them
	 char *aff_file = *argv++;
	 char *iso_file = *argv++;
	 compare_files(aff_file,iso_file);
	 exit(0);
     }
     usage();				// don't know what to do
     exit(0);
}
