/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */

#include "afflib.h"
#include "afflib_i.h"
#include <fcntl.h>
#include <assert.h>
#include <openssl/rand.h>
#include <errno.h>

/* Utility Functions */



/* Delete the first occurance of the named segment */
int	af_del_seg(AFFILE *af,const char *segname)
{
    size_t segsize;
    if(af_find_seg(af,segname,0,0,&segsize)!=0){
	return -1;
    }
    /* Now wipe it out */
    size_t ignore_size = segsize-strlen(segname);
    af_write_ignore(af,ignore_size);
    return 0;
}

/****************************************************************
 *** GET FUNCTIONS
 ****************************************************************/

/****************************************************************
 *** Probe the next segment:
 *** Return its name and argument, but don't advance the pointer....
 *** Returns 0 on success, -1 on end of file or other error.
 ****************************************************************/
/****************************************************************
 *** Get the next segment.
 *** data must not be freed.
 *** *datalen_ is size of data; *datalen_ is modified
 *** to amount of segment
 *** if no data is specified, then just seek past it...
 *** Returns 0 on success, 
 *** -1 on end of file.
 *** -2 if *data is not large enough to hold the segment
 ****************************************************************/
int af_get_next_seg(AFFILE *af,char *segname,size_t segname_len,unsigned long *arg,
			unsigned char *data,size_t *datalen_)
{
    if(!af->aseg) err(1,"af_get_next_segv only works with aff files");

    size_t data_len;
    if(af_probe_next_seg(af,segname,segname_len,arg,&data_len,0,0)){
	return -1;			// must be end of file or something
    }
    if(data){
	/* Allocate the data and read... */
	if(datalen_ == 0) errx(-2,"af_get_next_seg: data provided but datalen is NULL");
	if(data_len > *datalen_){
	    fseeko(af->aseg,data_len,SEEK_CUR);	// go back
	    fprintf(stderr,"af_get_next_seg: *datalen_=%d data_len=%d\n",(int)*datalen_,(int)data_len);
	    return -2;
	}
	if(fread(data,1,data_len,af->aseg)!=data_len){
	    fprintf(stderr,"af_get_next_segv: EOF on reading segment? File is corrupt.\n");
	    exit(1);
	}
    } else {
	fseeko(af->aseg,data_len,SEEK_CUR); // skip past the data
    }
    if(datalen_) *datalen_ = data_len;

    /* Now read the tail */
    struct af_segment_tail segt;
    if(fread(&segt,sizeof(segt),1,af->aseg)!=1){
	fprintf(stderr,"af_get_next_segv: end of file reading segment tail...\n");
	exit(1);
    }
    /* Validate tail */
    unsigned long stl = ntohl(segt.segment_len);
    unsigned long calculated_segment_len =
	sizeof(struct af_segment_head)
	+ strlen(segname)
	+ data_len + sizeof(struct af_segment_tail);

    if(strcmp(segt.magic,AF_SEGTAIL)!=0){
	fprintf(stderr,"af_get_next_segv: AF file is corrupt. No tail!\n");
	exit(1);
    }
    if(stl != calculated_segment_len){
	fprintf(stderr,"af_get_next_segv: AF file is corrupt reading tail (%lu!=%lu)!\n",
		stl,calculated_segment_len);
	exit(1);
    }
    return 0;
}



/****************************************************************
 *** AFF creation functions
 ****************************************************************/






