/*
 * afflib_db.cpp:
 * 
 * Functions for the manipulation of the AFF database.
 */

#include "afflib.h"
#include "afflib_i.h"


/****************************************************************
 *** Low-level functions
 ****************************************************************/

int af_rewindi(AFFILE *af)
{
    fseeko(af->aseg,sizeof(struct af_head),SEEK_SET); // go to the beginning
    return 0;
}


/* Are we at the end of the file?
 * apparently feof doesn't work.
 * We need to see if the pointer is literally at the end of the file...
 */
int af_seg_eof(AFFILE *af)
{
    int64 start = ftello(af->aseg);
    fseek(af->aseg,0L,SEEK_END);	// go to the end
    int64 end = ftello(af->aseg);
    fseek(af->aseg,start,SEEK_SET);	// go to where we started

    return start==end;			// if we were at the end, we are at the end
}

/****************************************************************
 *** Probe Functions 
 ****************************************************************/

int af_probe_next_seg(AFFILE *af,
		       char *segname,
		       size_t segname_len,
		       unsigned long *arg_, // optional arg
		       size_t *datasize_, // optional get datasize
		       size_t *segsize_, // optional get size of entire segment
		       int do_rewind) // optional rewind af->aseg, otherwise leave at data
{
    if(!af->aseg) err(1,"af_probe_segment only works with aff files");    

    struct af_segment_head segh = {0,0};

    uint64 start = ftello(af->aseg);

    if(fread(&segh,sizeof(segh),1,af->aseg)!=1){
	return -1;
    }
    if(strcmp(segh.magic,AF_SEGHEAD)!=0){
	err(1,"afflib: segh is corrupt at %qd\n",start);
    }

    unsigned long name_len = ntohl(segh.name_len);
    unsigned long datasize = ntohl(segh.data_len);
    if(name_len>AF_MAX_NAME_LEN){
	err(1,"afflib: name_len=%lu (an outrageous value)",name_len);
    }

    if(name_len+1 > segname_len){
	fseeko(af->aseg,start,SEEK_SET); // rewind to start
	return -2;
    }

    if(fread(segname,1,name_len,af->aseg)!=name_len){
	fseeko(af->aseg,start,SEEK_SET); // rewind
	return -1;
    }
    segname[name_len] = 0;

    if(do_rewind) fseeko(af->aseg,start,SEEK_SET); // rewind

    unsigned long segsize = sizeof(struct af_segment_head)
	+ sizeof(struct af_segment_tail)
	+ name_len + datasize;

    if(arg_)      *arg_      = ntohl(segh.flag);
    if(datasize_) *datasize_ = datasize;
    if(segsize_)  *segsize_ = segsize;

#ifdef DEBUG
    fprintf(stderr,"af_probe_next_seg(segname=%s datasize=%d segsize=%d) do_rewind=%d\n",
	    segname,datasize,segsize,do_rewind);
#endif
    af->last_command = 1;
    return 0;
}


/* af_backspace:
 * moves back one segment in the aff file
 */
int af_backspace(AFFILE *af)
{
    struct af_segment_tail segt;

    uint64 start = ftello(af->aseg);
    uint64 pos_tail = start - sizeof(segt); // backspace to read the tail
    fseeko(af->aseg,pos_tail,SEEK_SET);
    if(fread(&segt,sizeof(segt),1,af->aseg)!=1){
	fseeko(af->aseg,start,SEEK_SET); // put it back
	return -1;			// can't read segt?
    }
    /* Now I know how long the segment was. Compute where it started */
    uint64 seg_start = start - ntohl(segt.segment_len);
    fseeko(af->aseg,seg_start,SEEK_SET);
    return 0;
}



/* find the given segment and return 0 if found.
 * Leave the file pointer positioned at the start of the segment.
 * Return -1 if segment is not found, and leave pointer at the end
 */
int	af_find_seg(AFFILE *af,const char *segname,
		    unsigned long *arg,
		    size_t *datasize,
		    size_t *segsize)
{
    af_rewindi(af);

    char   next_segment_name[AF_MAX_NAME_LEN];
    size_t next_segsize = 0;
    size_t next_datasize = 0;
    unsigned long next_arg;
    
    while(af_probe_next_seg(af,next_segment_name,sizeof(next_segment_name),
			    &next_arg,&next_datasize,&next_segsize,1)==0){
	if(strcmp(next_segment_name,segname)==0){	// found the segment!
	    if(datasize) *datasize = next_datasize;
	    if(segsize)  *segsize  = next_segsize;
	    if(arg)      *arg      = next_arg;
	    return 0;			// return the info
	}
	fseeko(af->aseg,next_segsize,SEEK_CUR);	// skip the segment
    }
    return -1;				// couldn't find segment
}
		    
/****************************************************************
 *** low-level routines for reading 
 ****************************************************************/

/* af_get_segment:
 * Get the named segment.
 * If it is the next segment, just get it.
 * Otherwise go to the beginning and scan for it.
 * Eventually we hope to cache the names and starting locations of each segment...
 */

/* Fix this to use the directory */

int af_get_seg(AFFILE *af,const char *name,unsigned long *arg,
		unsigned char *data,size_t *datalen)
{
    char next[AF_MAX_NAME_LEN];
    int first = 1;
    size_t segsize = 0;
    
    do {

	if(af_probe_next_seg(af,next,sizeof(next),0,0,&segsize,1)==0){	
	    if(strcmp(next,name)==0){	// found the segment!
		int ret = af_get_next_seg(af,next,sizeof(next),arg,data,datalen);
		assert(strcmp(next,name)==0);	// hopefully it's still the same!
		return ret;
	    }
	}
	else {
	    /* Must be at EOF. If this is not first time through,
	     * then segment does not exist...
	     */
	    if(!first) break;
	}
	if(first){
	    /* If first time we checked, rewind to the first segment */
	    af_rewindi(af);
	    first = 0;			// no longer the first time through
	    continue;
	}
	fseeko(af->aseg,segsize,SEEK_CUR); // seek to next segment
    } while(1);
    return -1;				// couldn't find segment
}

int af_get_segq(AFFILE *af,const char *name,uint64 *aff_quad)
{
    unsigned char buf[8];
    size_t  bufsize = sizeof(buf);

    if(af_get_seg(af,name,0,(unsigned char *)&buf,&bufsize)){
	return -1;			// couldn't get it...
    }
    if(bufsize!=sizeof(struct aff_quad)){		// make sure size is good.
	return -1;
    }
	
    *aff_quad = af_decode_q(buf);
    return 0;
}








/****************************************************************
 *** low-level functions for writing
 ****************************************************************/

/* af_write_ignore:
 * write an AF_IGNORE structure of a given size.
 */

int af_ignore_overhead()
{
    return sizeof(struct af_segment_head)+sizeof(struct af_segment_tail);
}

int af_write_ignore(AFFILE *af,unsigned long bytes)
{
    unsigned char *invalidate_data = (unsigned char *)calloc(bytes,1);

    af_write_seg(af,AF_IGNORE,0,invalidate_data,bytes); // overwrite with NULLs
    free(invalidate_data);
    return(0);
}



/* af_write_seg:
 * put the given named segment at the current position in the file.
 * Return 0 for success, -1 for failure (probably disk full?)
 * TODO: Remove the incomplete transaction when writing...
 */

int af_write_seg(AFFILE *af, const char *name,unsigned long arg,const void *data,unsigned int datalen)
{
    struct af_segment_head segh;
    struct af_segment_tail segt;

    if(af->debug){
	fprintf(stderr,"af_write_seg(" POINTER_FMT ",'%s',%lu,data=" POINTER_FMT ",datalen=%u)\n",
		af,name,arg,data,datalen);
    }

    assert(sizeof(segh)==16);
    assert(sizeof(segt)==8);

    /* If the last command was not a probe (so we know where we are), and
     * we are not at the end of the file, something is very wrong.
     */

    unsigned int namelen = strlen(name);

    strcpy(segh.magic,AF_SEGHEAD);
    segh.name_len = htonl(namelen);
    segh.data_len = htonl(datalen);
    segh.flag      = htonl(arg);

    strcpy(segt.magic,AF_SEGTAIL);
    segt.segment_len = htonl(sizeof(segh)+namelen + datalen + sizeof(segt));
    
#ifdef DEBUG   
    {
	uint64 o = ftello(af->aseg);
	fprintf(stderr,"af_write_segv: putting segment %s (datalen=%d) offset=%qd\n",
		name,datalen,o);
    }
#endif

    if(fwrite(&segh,sizeof(segh),1,af->aseg)!=1) return -1;
    if(fwrite(name,1,namelen,af->aseg)!=namelen) return -1;
    if(fwrite(data,1,datalen,af->aseg)!=datalen) return -1;
    if(fwrite(&segt,sizeof(segt),1,af->aseg)!=1) return -1;
    fflush(af->aseg);			// make sure it is on the disk
    af->last_command = 2;
    return 0;
}



/* af_write_segq:
 * Put the named quad-byte valud at the current position in the file.
 */

int af_write_segq(AFFILE *af,const char *name,uint64 value)
{
    struct aff_quad  q;

    assert(sizeof(q)==8);

    q.low  = htonl((unsigned long)(value & 0xffffffff));
    q.high = htonl((unsigned long)(value >> 32));
    
    return af_write_seg(af,name,AF_SEG_QUADWORD,&q,8);
}

/****************************************************************
 *** Update functions
 ****************************************************************/

/*
 * af_update_segv:
 * Update the given named segment with the new value.
 * if "append" is false, don't append if it can't be found. 
 * if "append" is true and the named segment can't be found, append it.
 */

int af_update_seg(AFFILE *af, const char *name,
		    unsigned long arg,const void *value,unsigned int vallen,int append)
{
    char   next_segment_name[AF_MAX_NAME_LEN];
    size_t next_segsize = 0;
    size_t next_datasize = 0;

    /* if we are updating with a different size,
     * remember the location and size of the AF_IGNORE segment that
     * has the smallest size that is >= strlen(name)+vallen
     */
    size_t size_needed = strlen(name)+vallen+af_ignore_overhead();
    size_t size_closest = 0;
    uint64         loc_closest = 0;
       
#ifdef DEBUG
    fprintf(stderr,"af_update_seg(name=%s,arg=%lu,vallen=%u,append=%d)\n",name,arg,vallen,append);
#endif

    /* Is the last segment an ignore? */

    af_rewindi(af);			// start at the beginning

    while(af_probe_next_seg(af,next_segment_name,sizeof(next_segment_name),
			    0,&next_datasize,&next_segsize,1)==0){
	uint64 next_segment_loc = ftello(af->aseg);

#ifdef DEBUG
	fprintf(stderr,"  next_segment_name=%s next_datasize=%d "
		"next_segsize=%d next_segment_loc=%qd\n",
		next_segment_name,
		next_datasize,
		next_segsize,next_segment_loc);
#endif

	if(strcmp(next_segment_name,name)==0){	// found the segment!

	    /* Okay. Is there room to fit it here? */
	    if(next_datasize == vallen){
		/* We can just write in place! */
		int r = af_write_seg(af,name,arg,value,vallen);
		return r;
	    }

	    /* Invalidate this segment and continue scanning */
	    
	    uint64 loc_invalidated = next_segment_loc; // remember where this is

	    af_write_ignore(af,next_datasize+strlen(name));

	    uint64 loc_before_seek = ftello(af->aseg); // before the seek

	    fseeko(af->aseg,(uint64)0,SEEK_END);              // go to the end of the file

	    uint64 pos_eof = ftello(af->aseg);


	    if(pos_eof == loc_before_seek){
		/* Apparently I didn't move at all with the seek.
		 * If the segment being written is larger than the segment being replaced,
		 * we can just backspace.
		 */
		if(next_segsize < vallen){
		    //fprintf(stderr,"  *** seeking to %qd\n",loc_invalidated);
		    fseeko(af->aseg,loc_invalidated,SEEK_SET);
		}
	    }

	    append = 1;			// be sure to add the data to the end!
	    break;			// and exit this loop
	}

	/* If this is an AF_IGNORE, see if it is a close match */
	if(next_segment_name[0]==AF_IGNORE[0]){
	    if(next_datasize>=size_needed && ((next_datasize<size_closest || size_closest==0))){
		size_closest = next_datasize;
		loc_closest  = next_segment_loc;
	    }
	}
	fseeko(af->aseg,next_segsize,SEEK_CUR); // skip this segment
    }
    if(append){				// if appending, just write to the end
	/* See if there is a place where this would fit. */
	//fprintf(stderr,"size_closest=%d\n",size_closest);
	if(size_closest>0){
	    /* Yes. Put it here and put a new AF_IGNORE in the space left-over
	     * TODO: If the following space is also an AF_IGNORE, then combine the two.
	     */
#ifdef DEBUG
	    fprintf(stderr,"size_closest=%d\n",size_closest);
	    fprintf(stderr,"loc_closest=%qd\n",loc_closest);
	    fprintf(stderr,"*** Squeezing it in at %qd. name=%s. vallen=%d\n",loc_closest,name,vallen);
#endif

	    fseeko(af->aseg,loc_closest,SEEK_SET); // move to the location
	    af_write_seg(af,name,arg,value,vallen); // write the new segment

	    int newsize = size_closest - vallen - af_ignore_overhead() - strlen(name);

	    af_write_ignore(af,newsize); // write the smaller ignore
	    return 0;
	}
	return af_write_seg(af,name,arg,value,vallen);
    }
    return -2;				// couldn't find segment
}


/* af_update_segq:
 * Update the named aff_quad-byte value.
 */

int af_update_segq(AFFILE *af, const char *name, uint64 value,int append)
{
    struct aff_quad  q;
    q.low  = htonl((unsigned long)(value & 0xffffffff));
    q.high = htonl((unsigned long)(value >> 32));
    
    return af_update_seg(af,name,0,&q,8,append);
}


