/*
 * afflib_i.h:
 * The "master include file" of the AFF Library.
 * Includes many fucntions that are not designed to be used by application programmers.
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */

#ifndef AFFLIB_I_H
#define AFFLIB_I_H

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <zlib.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <assert.h>
#include <openssl/rand.h>
#include <errno.h>

#ifdef WIN32
#include "unix4win32.h"
#include <winsock.h>			// htonl()
#endif

#ifdef UNIX
#include <arpa/inet.h>
#include <err.h>
#include <unistd.h>
#endif

#ifndef O_BINARY
#define O_BINARY 0			// for Windows compatability
#endif

#ifdef __APPLE__
#define POINTER_FMT "%p"
#endif

#ifdef linux
#define POINTER_FMT "%p"
#endif

#ifndef POINTER_FMT
#define POINTER_FMT "%x"		// guess
#endif



/* The header for an AFF file. All binary numbers are stored in network byte order. */
#define AF_HEADER "AFF10\r\n\000"
struct af_head {
    char header[8];			// "AFF10\r\n\000"
    /* segments follow */
};


/* The header of each segment */
#define AF_SEGHEAD "AFF\000"
struct af_segment_head {
    char magic[4];			// "AFF\000"
    unsigned long name_len:32;		// length of segment name
    unsigned long data_len:32;		// length of segment data, if any
    unsigned long flag:32;		// argument for name;
    /* name follows, then data */
};

/* The tail of each segment */
#define AF_SEGTAIL "ATT\000"
struct af_segment_tail {
    char magic[4];			// "ATT\000"
    unsigned long segment_len:32;      // includes head, tail, name & length
};


/* How 64-bit values are stored in a segment */
struct aff_quad {
    unsigned long low:32;
    unsigned long high:32;
};

/* The as-yet-unimplemented AFF directory */
struct af_dir {
    int len;				// length of directory entry, including name and header
    struct aff_quad offset;		// offset, stored as an aff_quad
    char name[1];			// name of the entry; can be longer than 1;
					// does not include \000
};


/****************************************************************
 *** Not AFF functions at all, but placed here for convenience.
 ****************************************************************/
 
const char *af_hexbuf(char *dst,int dst_len,const unsigned char *bin,int bytes,int add_spaces);

/****************************************************************
 *** Lowest-level routines for manipulating the AFF File...
 ****************************************************************/


int	af_rewindi(AFFILE *af);		// go to the beginning of the AF file; doesn't move
                                        // the file pointer; use af_seek(af,0,0);

/* Navigating within the AFFILE */
/* probe the next segment.
 * Returns: 0 if success
 *          -1 if error
 *          -2 if segname_len was not large enough to hold segname
 *         - segname - the name of the next segment.
 *         - segsize - number of bytes the entire segment is.
 *      
 * doesn't change af->aseg pointer if do_rewind is true, otherwise leaves stream
 *           positioned ready to read the data
 */

int	af_probe_next_seg(AFFILE *af,char *segname,size_t segname_len,
			   unsigned long *arg,
			   size_t *datasize, size_t *segsize,
			   int do_rewind);
int	af_backspace(AFFILE *af);	// back up one segment



/* find the given segment and return 0 if found, filling in the fields.
 * Leave the file pointer positioned at the start of the segment.
 * Return -1 if segment is not found, and leave pointer at the end
 */
int	af_find_seg(AFFILE *af,const char *segname,
		    unsigned long *arg, size_t *datasize, size_t *segsize);	
		    

int	af_get_seg(AFFILE *af,const char *name,unsigned long *arg,
		    unsigned char *data,size_t *datalen);

/****************************************************************
 *** Writing functions
 ****************************************************************/


/* Low-level functions for putting
 * af_put_segv --- put a segment with an argument (also useful for putting numbers)
 * af_put_segq --- put a quadword, useful for putting 8-byte values.
 */

int	af_ignore_overhead();	
int     af_write_ignore(AFFILE *af,unsigned long bytes);
int	af_write_segq(AFFILE *af,const char *name,uint64 value);

int	af_write_seg(AFFILE *af,const char *name,unsigned long arg,
		     const void *value,unsigned int vallen); 

int	af_make_badflag(AFFILE *af);	// creates a badflag and puts it



/* afflib_stream.cpp
 * Mostly internal functions.
 */
void	af_set_dataseg_sizes(AFFILE *af,long segsize,uint64 imagesize);
int	af_update_image_seg(AFFILE *af,int segnum,unsigned char *data,int datalen);
int	af_purge(AFFILE *af);		// throw out the current data segment in memory

void	af_get_segsizes(AFFILE *af);	// sets up values if we can get them.
void	af_demand_segsizes(AFFILE *af);	// abort if we can't get them

/* afflib_util.cpp
 */
unsigned long long af_decode_q(unsigned char buf[8]); // return buf[8] into an unsigned quad

 



#endif
