/*
 * The AFFLIB data stream interface.
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */


#include "afflib.h"
#include "afflib_i.h"


/****************************************************************
 *** Internal Functions.
 ****************************************************************/

/* af_get_segsizes:
 * Get the segment sizes if they are set in the file.
 */
void af_get_segsizes(AFFILE *af)
{
    af_get_seg(af,AF_SEGSIZE,&af->image_datasegsize,0,0);
    af_get_segq(af,AF_IMAGESIZE,&af->imagesize);
    af->imagesize_orig = af->imagesize;
    if(af->segbuf==0 && af->image_datasegsize!=0){
	af->segbuf = (unsigned char *)malloc(af->image_datasegsize);
    }
}


/* af_demand_segsizes:
 * Get the segment sizes and fail if we can't get them...
 */
void af_demand_segsizes(AFFILE *af)
{
    if(af->image_datasegsize==0){
	af_get_segsizes(af);
    }
    if(af->image_datasegsize==0){
	errx(1,"af_demand_segsizes: af->image_datasegsize not set. Aborting.\n");
    }
}


/* af_get_image_segment:
 * Get a segment. If it is compressed, uncompress it.
 * data points to a segmenet of at least *bytes;
 * *bytes is then modified to indicate the actual amount of bytes read.
 * if shouldfree is set, then data should be freed.
 * Return 0 if success, -1 if fail.
 */

int af_get_image_segment(AFFILE *af,int segnum,
			 unsigned char *data,size_t *bytes)
{
    char segname[256];
    unsigned long arg=0;
    size_t segdata_len=0;

    /* Figure out the segment name from the number */
    sprintf(segname,AF_SEG,segnum);
    
    /* Find out the size of the segment and if it is compressed or not */
    if(af_find_seg(af,segname,&arg,&segdata_len,0)){
	return -1;			// segment doesn't exist
    }
       
    /* If the segment isn't compressed, just get it and return */
    if((arg & AF_SEG_COMPRESSED)==0){
	return af_get_seg(af,segname,0,data,bytes);
    }

    /* Allocate memory to hold the compressed segment */
    unsigned char *compressed_data = (unsigned char *)malloc(segdata_len);
    size_t compressed_data_len = segdata_len;
    if(compressed_data==0){
	return -2;			// memory error
    }

    /* Get the data */
    if(af_get_seg(af,segname,0,compressed_data,&compressed_data_len)){
	free(compressed_data);
	return -3;			// read error
    }

    /* Now uncompress into the user-provided buffer */
    uLongf bytes_ulf = *bytes;
    int r = uncompress(data,&bytes_ulf,compressed_data,compressed_data_len);
    *bytes = bytes_ulf;
    switch(r){
    case Z_OK:
	free(compressed_data);		// don't need this one anymore
	return 0;
    case Z_ERRNO:
	err(1,"Z_ERRNOR decompressing segment %d",segnum);
    case Z_STREAM_ERROR:
	err(1,"Z_STREAM_ERROR decompressing segment %d",segnum);
    case Z_DATA_ERROR:
	err(2,"Z_DATA_ERROR decompressing segment %d",segnum);
    case Z_MEM_ERROR:
	err(1,"Z_MEM_ERROR decompressing segment %d",segnum);
    case Z_BUF_ERROR:
	err(2,"Z_BUF_ERROR decompressing segment %d",segnum);
    case Z_VERSION_ERROR:
	err(2,"Z_VERSION_ERROR decompressing segment %d",segnum);
    default:
	err(3,"uncompress returned an invalid value in get_segment");
    }
    return -4;				// should never get here!
}




/****************************************************************
 *** User-visible functions.
 ****************************************************************/

AFFILE *af_open(const char *filename,int flags,int mode)
{
    /* make sure things were compiled properly */
    assert(sizeof(struct af_head)==8);
    assert(sizeof(struct af_segment_head)==16);
    assert(sizeof(struct af_segment_tail)==8);

    if(flags & O_WRONLY) err(1,"af_open: flag O_WRONLY not supported");

    int exist = (access(filename,R_OK) == 0);	// does the file exist?
    int fd = open(filename,flags | O_BINARY,mode);
    if(fd<0) return 0;			// couldn't open

    AFFILE *af = (AFFILE *)calloc(sizeof(AFFILE),1);

    af->fname	= strdup(filename);	// remember file name
    af->segnum  = -1;			// no segment loaded

    int creating = flags & O_TRUNC ? 1 : 0; // are we creating? 
    struct stat sb;

    if(exist && !creating && af_identify(filename)!=1){
	/* It's not an AFF file. If the file is more than 0 bytes long,
	 * then pass through to raw routines...
	 */
	af->raw=fdopen(fd,"rb");
	
	if(stat(filename,&sb)){
	    err(1,"stat(%s): ",filename);
	}
	af->imagesize = sb.st_size;
	return af;
    }
	
    /* If we are creating, we must be creating an AFF... */
    char strflag[8];
    strcpy(strflag,"r");		// we have to be able to read
    if(flags & O_RDWR) 	strcpy(strflag,"w+");

    af->aseg = fdopen(fd,strflag);
    if(!af->aseg) err(1,"fdopen(%d,%s)",fd,strflag);

    /* If file is empty, then make it an AFF file! */
    if(fstat(fd,&sb)){
	err(1,"fstat(%s): ",filename);
    }
    if(sb.st_size==0){
	fwrite(AF_HEADER,1,8,af->aseg);	// writes the header 
	af_make_badflag(af);		// writes the flag for bad blocks
	return af;
    }
    
    /* We are opening an existing file. Verify once more than it is an AFF file
     * and skip past the header...*/

    /* Oh, we are reading. See if it is an AF header */
    char buf[8];
    if(fread(buf,sizeof(buf),1,af->aseg)!=1){
	/* Hm. End of file. That shouldn't happen here. */
	err(1,"af_open: couldn't read AFF header on existing file?");
    }

    if(strcmp(buf,AF_HEADER)!=0){
	buf[7] = 0;
	errx(1,"%s is not an AFF file (header=%s)\n",filename,buf);
    }
    af_get_segsizes(af);		// get the segment sizes if we can...
    return af;				// everything must be okay.
}

/* Open a regular file as an affile */
AFFILE *af_freopen(FILE *file)
{
    AFFILE *af = (AFFILE *)calloc(sizeof(AFFILE),1);
    af->raw = file;
    return af;
}

/* Open a regular file as an affile */
AFFILE *af_popen(const char *command,const char *type)
{
    if(strcmp(type,"r")!=0){
	errx(1,"only type 'r' supported on af_popen");
    }
    AFFILE *af = (AFFILE *)calloc(sizeof(AFFILE),1);
    af->raw = popen(command,"r");
    af->raw_popen = 1;
    return af;
}

/* Close the image */
int af_close(AFFILE *af)
{
    if(!af) return -1;			// invalid handle
    af_purge(af);			// empty
    if(af->writing){
	if(af->imagesize != af->imagesize_orig){
	    af_update_segq(af,AF_IMAGESIZE,af->imagesize,1);
	}
    }
    if(af->raw){
	if(af->raw_popen) pclose(af->raw);
	else fclose(af->raw);
    }
    if(af->aseg)  fclose(af->aseg);
    if(af->fname) free(af->fname);
    if(af->segbuf) free(af->segbuf);
    memset(af,0,sizeof(*af));
    free(af);
    return 0;
}


/* Return if we are at the end of the file */


int af_eof(AFFILE *af)
{
    if(af->raw) return feof(af->raw);
    if(af->image_datasegsize==0) af_demand_segsizes(af);
    return af->pos >= af->imagesize;
}

/* Write a actual data segment to the disk */
int af_update_image_seg(AFFILE *af,int segnum,unsigned char *data,int datalen)
{
    int ret = 0;
    unsigned int starting_csegs_written = af->csegs_written;
    char vbuf[32];
    sprintf(vbuf,AF_SEG,(int)segnum);

    if(af->compression_type != AF_COMPRESSION_NONE){
	/* Try to compress the data */
	uLongf destLen = af->image_datasegsize;	// it could be up to this big?
	unsigned char *cdata = 	cdata = (unsigned char *)malloc(destLen); 
	int cres = -1;			// compression result; default is that we failed
	
	if(cdata!=0){		// If data could be allocated
	    if(af->compression_callback) (*af->compression_callback)(0,0); // compression starting
	    cres = compress2((Bytef *)cdata, &destLen,
			     (Bytef *)data,datalen, af->compression_level);
	    if(af->compression_callback) (*af->compression_callback)(datalen, destLen);

	    if(cres==0 && destLen < af->image_datasegsize){
		/* Prepare to write out the compressed segment with compression */
		ret = af_update_seg(af,vbuf,AF_SEG_COMPRESSED,cdata,destLen,1);
		if(ret==0){
		    af->csegs_written++;
		}
	    }
	    free(cdata);
	    cdata = 0;
	}
    }
    
    /* If a compressed segment was not written, write it uncompressed */
    if(af->csegs_written == starting_csegs_written){ 
	ret = af_update_seg(af,vbuf,0,data,datalen,1);
	if(ret==0){
	    af->segs_written++;
	}
    }
    return ret;
}

/* Throw out the current segment */
int af_purge(AFFILE *af)
{
    int ret = 0;

    /* If the segment is dirty, we need to write it back
     * (and, if there is already a segment, we need to invalidate it.)
     */
    if(af->dirty){
	ret = af_update_image_seg(af,af->segnum,af->segbuf,af->image_datasegsize);
	af->dirty = 0;
    }
    af->segnum = -1;			// don't trust it anymore
    return ret;
}


/* Read from the affile */
int af_read(AFFILE *af,unsigned char *buf,int count)
{
    if(af->raw){	/* passthrough? */
	return fread(buf,1,count,af->raw);
    }

    af_demand_segsizes(af);    /* We need to have a datasegsize to continue */

    int total = 0;
    while(count>0){
	/* If no segment is loaded, load the current segment */
	if(af->segnum<0){
	    af->segnum = af->pos / af->image_datasegsize;	// will be the segment we want
	    af->segbytes = af->image_datasegsize;		// we can hold this much
	    if(af_get_image_segment(af,af->segnum,af->segbuf, &af->segbytes)){
		break;			// no more to get
	    }
	}
	// Compute how many bytes can be copied...
	// where we were reading from
	unsigned int seg_offset = af->pos - af->segnum * af->image_datasegsize; 
	unsigned int seg_left   = af->segbytes - seg_offset; // number we can get out
	unsigned int bytes_to_read = count;

	if(bytes_to_read > seg_left) bytes_to_read = seg_left;
	if(bytes_to_read > af->imagesize - af->pos) bytes_to_read = af->imagesize - af->pos;

	assert(bytes_to_read >= 0);	// 
	if(bytes_to_read==0) return total; // that's all we could get

	/* Copy out the bytes for the user */
	memcpy(buf,af->segbuf+seg_offset,bytes_to_read); // copy out
	buf     += bytes_to_read;
	af->pos += bytes_to_read;
	count   -= bytes_to_read;
	total   += bytes_to_read;

	/* If we read out all of the bytes that were left in the segment,
	 * then we are at the end of the segment, so throw it away.
	 */
	if(seg_left == bytes_to_read){	
	    af_purge(af);
	}
    }
    /* We have copied all of the user's requested data, so return */
    return total;
}

int af_write(AFFILE *af,unsigned char *buf,int count)
{
    if(!af->writing){
	errno = EPERM;			// operation not permitted
	return -1;			// not opened for writing
    }
    af_demand_segsizes(af);

    int total = 0;
    while(count>0){
	/* If no segment is loaded, load the current segment */
	if(af->segnum<0){
	    af->segnum = af->pos / af->image_datasegsize;	// will be the segment we want
	    af->segbytes = af->image_datasegsize;
	    if(af_get_image_segment(af,af->segnum,af->segbuf, &af->segbytes)){
		/* Create a new segment! */
		memset((char *)af->segbuf,'E',af->image_datasegsize);
		af->segbytes = 0;

		/* Fill the segment with the bad block flag... */
		unsigned char *cc;
		int i=0;
		for(cc=af->segbuf;cc<af->segbuf+af->image_datasegsize;cc++){
		    *cc = af->badflag[i++];
		    if(i>AFF_SECTOR_SIZE) i=0;
		}
	    }
	}
	unsigned int seg_offset = af->pos - af->segnum * af->image_datasegsize; // where we were writing to
	unsigned int seg_left   = af->image_datasegsize - seg_offset; // number we can write into
	unsigned int bytes_to_write = count;

	if(bytes_to_write > seg_left) bytes_to_write = seg_left;

	assert(bytes_to_write >= 0);	// 
	if(bytes_to_write==0) return total; // that's all we could get

	/* Copy out the bytes for the user */
	memcpy(af->segbuf+seg_offset,buf,bytes_to_write); // copy out

	if(af->segbytes < seg_offset+bytes_to_write){
	    af->segbytes = seg_offset+bytes_to_write; // it has been extended.
	}

	buf     += bytes_to_write;
	af->pos += bytes_to_write;
	count   -= bytes_to_write;
	total   += bytes_to_write;
	af->dirty = 1;

	/* If we wrote out all of the bytes that were left in the segment,
	 * then we are at the end of the segment, write it back...
	 */
	if(seg_left == bytes_to_write){	
	    if(af_purge(af)) return -1;
	}

	/* If we have written more than the image size, update the image size */
	if(af->pos > af->imagesize){
	    af->imagesize = af->pos;
	}

    }
    /* We have copied all of the user's requested data, so return */
    return total;
}



/* Seek in the virtual file */
uint64 af_seek(AFFILE *af,uint64 pos,int whence)
{
    uint64 new_pos=0;

    if(af->raw) return fseeko(af->raw,pos,whence);
    af_demand_segsizes(af);

    switch(whence){
    case SEEK_SET:
	new_pos = pos;
	break;
    case SEEK_CUR:
	new_pos += pos;
	break;
    case SEEK_END:
	new_pos = af->imagesize - pos;
	break;
    }
    if(new_pos < 0) new_pos = 0;	   // new change
    if(af->pos == new_pos) return af->pos; // no change
    af->pos = new_pos;			// set the new position

    /* If the correct segment is not loaded, purge the segment */
    int new_seg = af->pos / af->image_datasegsize;
    if(new_seg != af->segnum){
	af_purge(af);
    }

    return af->pos;
}

uint64  af_tell(AFFILE *af)
{
    if(af->raw) return ftello(af->raw);
    return af->pos;
}


/****************************************************************
 *** Misc functions
 ****************************************************************/


int af_enable_writing(AFFILE *af,int flag)
{
    int old_flag = af->writing;
    af->writing = flag;
    af->compression_callback = 0;	
    af->compression_type     = AF_COMPRESSION_ZLIB;
    af->compression_level    = Z_DEFAULT_COMPRESSION;
    return old_flag;
}

void af_enable_compression(AFFILE *af,int type,int level,void (*cb)(int,int))
{
    af->compression_type  = type; 
    af->compression_level = level;
    af->compression_callback = cb;
}

const char *af_filename(AFFILE *af)
{
    return af->fname;
}

const unsigned char *af_badflag(AFFILE *af)
{
    return af->badflag;
}

int af_make_badflag(AFFILE *af)
{
    RAND_pseudo_bytes(af->badflag,sizeof(af->badflag));
    strcpy((char *)af->badflag,"BAD SECTOR");

    if(af_write_seg(af,AF_BADFLAG,0,af->badflag,sizeof(af->badflag))) return -1;
    if(af_write_seg(af,AF_BADSECTORS,0,0,0)) return -1;
    return 0;
}


/*
 * af_set_dataseg_sizes:
 * Sets the segment and image size.
 * Doesn't accept new datasegsize ones if they are already set...
 */
void af_set_dataseg_sizes(AFFILE *af,long data_segsize,uint64 imagesize)
{
    /* can't change the image_datasegsze once it is set */
    if(af->image_datasegsize==0 && data_segsize!=0){
	af_write_seg(af,AF_SEGSIZE,data_segsize,0,0);
	af->image_datasegsize	= data_segsize;
	af->segbuf = (unsigned char *)malloc(data_segsize);
    }

    af_write_segq(af,AF_IMAGESIZE,imagesize);
    af->imagesize		= imagesize;
    af->imagesize_orig		= imagesize;
}


