#include "afflib.h"
#include "afflib_i.h"
#include "imager.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <term.h>
#include <assert.h>
#include <errno.h>

/* figure_media():
 * This is a function that returns, for an open device, the sector_size, total_sectors,
 * and maximum number of blocks that can be read at a time (or 0 if there is no max).
 * There is a generic implementation at the bottom, and operating-specific versions
 * above.
 */

#ifdef __APPLE__
#include <sys/disk.h>
int figure_media(int fd,int *sector_size,uint64 *total_sectors,int *maxreadblocks)
{
    if(ioctl(fd,DKIOCGETBLOCKSIZE,sector_size)){
        fprintf(stderr,"ioctl(DKIOCGETBLOCKSIZE) %s --- assuming 512\n",
		strerror(errno));
	*sector_size = 512;
    }
    if(ioctl(fd,DKIOCGETBLOCKCOUNT,total_sectors)){
        fprintf(stderr,"ioctl(DKIOCGETBLOCKSIZE) %s --- seeking not allowed on input.\n",
		strerror(errno));
	*total_sectors=0;
    }
    if(ioctl(fd,DKIOCGETMAXBLOCKCOUNTREAD,maxreadblocks)){
	/* This really isn't an error. */
	*maxreadblocks = 0;
    }
    return 0;
}
#define MEDIA_FIGURED
#endif

#ifdef __FreeBSD__
#include <sys/disk.h>
int figure_media(int fd,int *sector_size,uint64 *total_sectors,int *maxreadblocks)
{
    /* Figure out how many bytes to convert if we are not faking it*/
    if(ioctl(fd,DIOCGSECTORSIZE,sector_size)){
        err(1,"ioctl(DIOCGSECTORSIZE)");
    }
    off_t inbytes;
    if(ioctl(fd,DIOCGMEDIASIZE,&inbytes)){
        err(1,"ioctl(DIOCGMEDIASIZE)");
    }
    if(inbytes % *sector_size != 0){
	fprintf(stderr,"ioctl(DIOCGSECTORSIZE) returns %d bytes\n",*sector_size);
	fprintf(stderr,"ioctl(DIOCGMEDIASIZE) returns %d bytes\n",inbytes);
	fprintf(stderr,"which is not an even number of sectors. Something is wrong.\n");
	exit(1);
    }
    *total_sectors = inbytes / *sector_size;
    *maxreadblocks = 0;
    return 0;
}
#define MEDIA_FIGURED
#endif

#ifdef linux
#include <linux/fs.h>
#include <sys/ioctl.h>
int figure_media(int fd,int *sector_size,uint64 *total_sectors,int *maxreadblocks)
{
    /* Figure out how many bytes to convert if we are not faking it*/
  uint64 total_bytes;
  *sector_size = BLOCK_SIZE;	
    if(ioctl(fd,BLKGETSIZE64,&total_bytes)){
        err(1,"ioctl(BLKGESIZE64)");
    }
    *total_sectors = total_bytes / *sector_size;
    *maxreadblocks = 0;
    return 0;
}
#define MEDIA_FIGURED
#endif


/* Okay. We can't figure it out. Try our best-effort guess,
 */

#ifndef MEDIA_FIGURED
int figure_media(int fd,int *sector_size,uint64 *total_sectors,int *maxreadblocks)
{
    *sector_size = 512;			// it's a good guess

    /* Let's try seeking to the end of fd and ask where we are! */
    
    off_t end_of_file = lseek(fd,0,SEEK_END);

    if(end_of_file==-1){
	end_of_file = 0;		// can't figure it.
    }

    lseek(fd,0,SEEK_SET);		// go back to the beginning

    *total_sectors = end_of_file / *sector_size;
    *maxreadblocks = 0;
    return 0;
}
#endif



/*
 * imager.cpp:
 * The C++ imaging library.
 */

imager::imager()
{
    last_sector_read = 0;	// sector number
    bad_sectors_read = 0;
    total_sectors_written = 0;
    total_sectors_read=0;
    total_bytes_read = 0;
    total_bytes_written = 0;
    total_blank_sectors = 0;
    af = 0;
    fout = 0;
    status = 0;
    hash_invalid = false;
    imaging = false;

    /* error recovery */
    error_recovery_phase = 0;
    errors_in_a_row = 0;
}


void imager::set_fout(FILE *fout_)
{
    fout = fout_;
}
    
/* Any additional AFF initialization */
void imager::set_af(AFFILE *af_)
{
    af = af_;

    /* Store empty hashes, which will be updated later. */
    
    char buf[256];
    memset(buf,0,sizeof(buf));
    af_write_seg(af,AF_MD5,0,buf,16);	// leave 16 bytes for MD5
    af_write_seg(af,AF_SHA1,0,buf,20);	// leave 20 bytes for SHA1
}

void imager::write_data(unsigned char *buf,uint64 offset,int len)
{
    if(offset != total_bytes_written){
	hash_invalid = true;		// oh well; we must have had a seek
    }
    else {
	/* Update hash functions. */
	MD5_Update(&md5,buf,len);
	SHA1_Update(&sha,buf,len);
    }

    /* Write it out and carry on... */
    if(af){
	af_seek(af,offset,SEEK_SET);
	if(af_write(af,buf,len)!=len){
	    perror("af_write");	// this is bad
	    af_close(af);	// try to gracefully recover
	    fprintf(stderr,"Imaging terminated because af_write failed.\n");
	    fprintf(stderr,"Correct error condition and re-run aimage. You should\n");
	    fprintf(stderr,"be able to carry on from where you left off.\n");
	    exit(1);
	}
    }
    if(fout){
	fseeko(fout,offset,SEEK_SET);
	if(fwrite(buf,1,len,fout)!=(unsigned)len){
	    perror("fwrite");	// this is also bad
	    fclose(fout);	// get what we can out to disk
	    fprintf(stderr,"Imaging terminated because fwrite() failed.\n");
	    fprintf(stderr,"Correct error condition and re-run aimage.\n");
	    fprintf(stderr,"You may be able to carry on from where you left off.\n");
	    exit(1);
	}
    }
    total_sectors_written += len/sector_size;
    total_bytes_written   += len;
}


void imager::image_loop(uint64 low_water_mark,
			uint64 high_water_mark,
			int direction, int readsectors,int error_mask)
{
    // buffer to store the data we read
    unsigned char *buf = (unsigned char *)malloc(readsectors*sector_size); 
    unsigned char *blank = (unsigned char *)malloc(sector_size);

    if(!buf) err(1,"malloc");
    if(!blank) err(1,"malloc");

    memset(blank,0,sector_size);

    /* Loop as long as we have room, or until we get an EOF
     * (if high_water_mark is 0.)
     */
    imaging = true;
    while(low_water_mark < high_water_mark || high_water_mark==0){ 

	/* Figure out where to read and how how many sectors to read */

	uint64 snum;	// where we will be reading
	int sectors_to_read = readsectors;
	if(sectors_to_read > maxreadblocks && maxreadblocks>0){
	    sectors_to_read = maxreadblocks;
	}
	if(direction==1){	// going up
	    snum = low_water_mark;

	    /* If a high water mark is set, take it into account */
	    if(high_water_mark>0){
		int sectors_left = high_water_mark - snum;
		if(sectors_left < sectors_to_read){ 
		    sectors_to_read = sectors_left;
		}
	    }
	}
	else {
	    assert(high_water_mark != 0); // we can't go backwards if we don't know end

	    snum = high_water_mark - sectors_to_read;
	    if(snum<low_water_mark){
		snum = low_water_mark;
		sectors_to_read = high_water_mark - low_water_mark;
	    }
	}

	last_sector_read = snum;
	last_sectors_read = sectors_to_read;
	last_direction = direction;

	if(status) (*status)();			// call the status function

	/* Now seek and read */
	reading.start();
	uint64 data_offset = sector_size * snum;
	lseek(in,data_offset,SEEK_SET); // make sure we are at the right place
	int bytes_to_read = sectors_to_read * sector_size;
	int bytes_read    = read(in,buf,bytes_to_read);
	reading.stop();

	if(bytes_read == bytes_to_read){
	    /* Got a good read! */
	    total_sectors_read    += sectors_to_read;
	    total_bytes_read      += bytes_read;
	    errors_in_a_row = 0;

	    /* Count the blank sectors */
	    for(int i=0;i<sectors_to_read;i++){
		if(memcmp(buf+i*sector_size,blank,sector_size)==0){
		    total_blank_sectors++;
		}
	    }
	    

	    /* Write the data! */


	    write_data(buf,data_offset,bytes_read);

	    if(direction==1){
		low_water_mark += sectors_to_read;
	    }
	    else {
		high_water_mark -= sectors_to_read;
	    }

	    continue;
	}

	/* If high water mark is 0,
	 * then just write out what we read and give up.
	 */
	if(high_water_mark==0){
	    write_data(buf,data_offset,bytes_read);
	    break;
	}

	/* Error handling follows */
	if(error_mask==0){
	    /* Fill buffer that wasn't read with bad block flag
	     * and write it out.
	     */
	    for(unsigned char *cc=buf;
		cc < buf+bytes_to_read-AFF_SECTOR_SIZE;
		cc += AFF_SECTOR_SIZE){
		memcpy(cc,af_badflag(af),AFF_SECTOR_SIZE);
	    }
	    write_data(buf,data_offset,bytes_to_read);
	    bad_sectors_read += sectors_to_read;
	    snum += sectors_to_read*direction;	// increment sector number counter
	    errors_in_a_row += 1;

	    /* See if the errors exceed the thresshold */
	    if(errors_in_a_row>retry_count){
		if(direction == -1){
		    /* That's it. Give up */
		    break;
		}
		errors_in_a_row = 0; // reset count
		direction = -1;
		continue;
	    }
	}
	if(error_mask==1){
	    /* Stop reading at the first error and write the incomplete buffer */
	    if(bytes_read>0){
		write_data(buf,data_offset,bytes_read);
	    }
	    break;
	}
    }
    imaging = false;
    free(blank);
    free(buf);
}

/* Returns 0 if okay, -1 if failure. */
int imager::set_input(int ifd)
{
    in = ifd;

    /* Make sure infile is actually a device, and not a file */
    struct stat so;
    if(fstat(ifd,&so)){
      perror("fstat");
    }
    int mode = so.st_mode & S_IFMT;

    /* Now figure out how many input blocks we have */
    if(mode==S_IFBLK || mode==S_IFCHR){
	figure_media(in,&sector_size,&total_sectors,&maxreadblocks);
	return 0;
    }
    if(mode==S_IFREG && allow_regular==false){
	fprintf(stderr,"input is a regular file.\n");
	fprintf(stderr,"Use aconvert to convert regular files to AFF.\n");
	return -1;
    }

    /* Okay. We don't know how big it will be, so just get what we can... */
    sector_size   = 512;		// it's a good guess
    total_sectors = 0;			// we don't know
    maxreadblocks = 0;			// no limit
    return 0;
}


void imager::hash_setup()
{
    /* Set up the MD5 & SHA1 machinery */
    MD5_Init(&md5);
    SHA1_Init(&sha);
}


float imager::compression_ratio()
{
    struct stat so;
    if(stat(af_filename(af),&so)){
	return -1.0;
    }
    return ((float)total_bytes_read-so.st_size)*100.0/total_bytes_read;

}
